from itertools import *
from piep import shell
import operator
from piep.line import Line
from .pycompat import *

class BaseList(object):
	'''
	Contains the common methods for :class:`piep.List` and :class:`piep.Stream`
	'''
	def divide(self, pred, keep_header=True):
		'''
		Divide this stream at lines where ``pred`` returns true.
		If ``keep_header`` is set to ``False``, lines matching ``pred`` will not
		be included in the results.

		Each group is returned as a :class:`List` of items.
		'''
		def it(src):
			group = []
			for item in src:
				if pred(item):
					if group:
						yield List(group)
					group = []
					if keep_header:
						group.append(item)
				else:
					group.append(item)
			if group:
				yield List(group)
		return self._replace(it(self.src))

	def map(self, fn):
		def _transform(line):
			result = fn(line)
			if isinstance(result, bool):
				return line if result else None
			return result
		return self._replace(filter(lambda x: x is not None, map(_transform, self.src)))

	def uniq(self, stable=False):
		'''Remove duplicates. Note: if ``stable`` is not given (or is ``False``),
		the order of the return value will be in an arbitrary order.
		If ``stable`` is ``True``, the order will be maintained and only the
		first occurance of a duplicate line will be kept. This is not the default
		because it's much slower. For sorted unique output, try ``sort(uniq=True)``'''
		items = self
		if stable:
			# element order preserved
			seen = {}
			items = [ seen.setdefault(x,x) for x in items if x not in seen ]
		else:
			items = set(items)
		return List(items)

	def zip(self, *others):
		'''Combine this stream with another, yielding sequential pairs from each stream.
		When one sequence is shorter than the other, it's padded with ``None`` elements.
		Basically, ``itertools.zip_longest(self, *others)``
		
		>>> list(Stream([1,2,3,4]).zip(['one','two','three']))
		[(1, 'one'), (2, 'two'), (3, 'three'), (4, None)]
		'''
		return self._replace(zip_longest(self.src, *others))

	def zip_shortest(self, *others):
		'''Like :data:`zip`, but stops once any of the sequences ends.
		
		>>> list(Stream([1,2,3,4]).zip_shortest(['one','two','three']))
		[(1, 'one'), (2, 'two'), (3, 'three')]
		'''
		return self._replace(zip(self.src, *others))

	def map_index(self, fn):
		i = [0]
		def _call(line):
			try:
				return fn(line, i[0])
			finally:
				i[0] += 1
		return self.map(_call)

	def filter(self, f=None):
		'''alias for itertools.ifilter(self, f)'''
		return self._replace(ifilter(f, self.src))

	def join(self, s):
		'''alias for ``s.join(self)``'''
		return s.join(list(self))

	def merge(self):
		'''
		Combine a sequence of iterables into one sequence.
		
		>>> list(Stream([[1],[2,3],[4,5]]).merge())
		[1, 2, 3, 4, 5]
		'''
		return self._replace(chain.from_iterable(self.src))

	def flatten(self):
		r'''
		Combine a sequence of strings (containing newlines) into a sequence of lines.
		
		>>> list(Stream(["a\nb\nc","d\ne"]).flatten())
		['a', 'b', 'c', 'd', 'e']
		'''
		return self._replace(chain.from_iterable(map(lambda x: Line(x).splitlines(), self.src)))

	def sort(self, uniq=False):
		'''
		Return a sorted version of this stream (note: reads entire stream into memory).
		Alias for ``sorted(self)``.

		When ``uniq``=``True``, duplicates are removed from the result.
		'''
		items = set(self) if uniq else self
		return self._replace(sorted(items))

	def sortby(self, fn=None, key=None, attr=None, method=None):
		'''
		Return a sorted version of this stream (note: reads entire stream into memory).
		One (and only one) of the argument types should be provided as the sort key:
		
		- ``fn`` will sort using the return value of calling ``fn`` with each item: ``fn(item)``
		- ``key`` will sort using the given key of each element: ``item[key]``
		- ``attr`` will sort using the given attribute of each element: ``item.attr``
		- ``method`` will sort using the result of calling the given method (with no arguments) on each element: ``item.method()``
		'''
		defined_items = list(filter(lambda x: x is not None, (fn, key, attr, method)))
		assert len(defined_items) == 1, "exactly one of (fn, key, attr, method) arguments allowed to `sortby` method (you gave %s: %r)" % (len(defined_items), defined_items)

		if key is not None: fn = operator.itemgetter(key)
		if attr is not None: fn = operator.attrgetter(attr)
		if method is not None: fn = operator.methodcaller(method)
		return self._replace(sorted(self, key=fn))

	def reverse(self):
		'''
		Return a reversed version of this stream (note: reads entire stream into memory)
		Alias for ``reversed(self)``
		'''
		return self._replace(reversed(self))


class List(list, BaseList):
	def __init__(self, *a, **k):
		super(List, self).__init__(*a,**k)
		self.src = self
	
	def _replace(self, src):
		return List(src)
	
	len = list.__len__

class _SafeBoolIterator(object):
	@classmethod
	def of_iter(cls, it):
		try:
			head = next(it)
		except StopIteration:
			return _EmptyIterator()
		else:
			return _NonEmptyIterator(head, it)

	def __iter__(self): return self

	# py2 compat
	def next(self): return self.__next__()
	def __nonzero__(self): return self.__bool__()

class _EmptyIterator(_SafeBoolIterator):
	def __next__(self): raise StopIteration
	def __bool__(self): return False

class _NonEmptyIterator(_SafeBoolIterator):
	"""
	An iterator which keeps a reference to its own `head`,
	useful to check the first element without actually
	consuming it.

	>>> it = _NonEmptyIterator(1, [2,3])
	>>> list(it)
	[1, 2, 3]
	"""
	def __init__(self, head, tail):
		self._replace(head, iter(tail))

	def _replace(self, head, tail):
		self._head = head
		self._tail = tail
		self._it = None

	def __bool__(self):
		if self._it is None:
			# head not consumed
			return True
		else:
			# head consumed, consume the next item (if any) and
			# keep it as the new head
			try:
				head = next(self._it)
			except StopIteration:
				return False
			else:
				self._replace(head, self._tail)

	def __iter__(self):
		return self

	def __next__(self):
		if self._it is None:
			self._it = self._tail
			return self._head
		else:
			return next(self._it)

class Stream(BaseList):
	def __init__(self, src):
		self._replace(src)
	
	def __getitem__(self, n):
		"""
		>>> Stream([1,2,3,4,5,6,7])[2]
		3
		>>> Stream([1,2,3,4,5,6,7])[-2]
		6
		>>> list(Stream([1,2,3,4,5,6,7])[3:5])
		[4, 5]
		>>> list(Stream([1,2,3,4,5,6,7])[3:-2])
		[4, 5]
		>>> list(Stream([1,2,3,4,5,6,7])[:-2])
		[1, 2, 3, 4, 5]
		>>> list(Stream([1,2,3,4,5,6,7])[-2:])
		[6, 7]
		>>> list(Stream([1,2,3,4,5,6,7])[-4:-5])
		[]
		>>> list(Stream([1,2,3,4,5,6,7])[-4:-2])
		[4, 5]
		>>> list(Stream([1,2])[-4:])
		[1, 2]
		>>> list(Stream([1])[:2])
		[1]
		"""
		if isinstance(n, slice):
			return self._slice(n.start, n.stop, n.step)
		try:
			if n < 0:
				return pad_end(abs(n), self.src).end[0]
			elif n > 0:
				drop(n, self.src)
			return next(self.src)
		except StopIteration:
			raise IndexError(n)

	def _replace(self, src):
		self.src = iter(src)
		return self
	
	# unfortunately, list(item) will call __len__
	# and therefore drain the iter if it's defined
	# (so we just have a `.length` method instead)
	def len(self):
		return iter_length(self.src)

	def __add__(self, other):
		return list(self) + other

	def __radd__(self, other):
		return other + list(self)

	def _slice(self, start, stop, step):
		start = start or 0
		step = step or 1

		negative_stop = stop is not None and stop < 0
		positive_stop = stop is not None and stop >= 0

		# START
		if start >= 0:
			if positive_stop or stop is None:
				# normal slice
				self._replace(islice(self.src, start, stop))
			else:
				# normal start to negative end, e.g lst[2:-1]
				self._replace(drop(start, self.src))
				self._replace(pad_end(abs(stop), self.src))

		else: # start < 0
			if negative_stop:
				# e.g lst[-5:-2]
				# get the padded end
				padded = pad_end(abs(stop), self.src)
				if start >= stop:
					self._replace([])
				else:
					start -= stop
					self._replace(pad_end(abs(start), padded).end)

			elif stop is None:
				self._replace(pad_end(abs(start), self.src).end)

			else: # stop is positive, e.g lst[-3,2]
				try:
					first_half = take(stop, self.src)
					num_remaining = iter_length(self.src)
					if num_remaining > abs(start):
						raise StopIteration() # not really, but the slice will be empty
				except StopIteration:
					self._replace([])
				else:
					start += num_remaining # we already "chopped" num_remaining off the end of the original sequence
					self._replace(first_half[start:])

		# implement stepping as a post-filter once we've organised the slice
		if step != 1:
			self._replace(islice(self.src, 0, None, step))
		return self
	
	def __reversed__(self):
		rev = reversed(list(self.src))
		self._replace(rev)
		return List(rev)

	def __iter__(self):
		return self.src

	def __str__(self):
		return Line("\n".join(list(self.src)))

	def __repr__(self):
		return repr(list(self.src))

	def __bool__(self):
		"""
		>>> bool(Stream([1,2,3]))
		True

		>>> bool(Stream([]))
		False

		>>> bool(Stream(repeat(1)))
		True

		# ensure `bool` doesn't consume `head`
		>>> s = Stream(iter([1]))
		>>> bool(s)
		True
		>>> bool(s)
		True
		>>> bool(s)
		True
		"""
		# force self.src to be a bool-safe iterator first
		if not isinstance(self.src, _SafeBoolIterator):
			self._replace(_SafeBoolIterator.of_iter(self.src))
		return bool(self.src)

	__nonzero__ = __bool__

class pad_end(object):
	"""
	Creates an iterator with a "padded end". e.g
	an iterator with 2 padding at the end will maintain
	a buffer of 2 items, and stop 2 items before
	the end of the underlying iterator.

	You can call `.end` to get the last `n` items as a list.
	NOTE: this will exhaust the underlying iterator
	first.

	>>> list(pad_end(2, iter([1,2,3,4])))
	[1, 2]
	>>> pad_end(2, iter([1,2,3,4])).end
	[3, 4]
	>>> pad_end(4, iter([1,2,3,4,5,6,7,8,9])).end
	[6, 7, 8, 9]
	"""
	def __init__(self, n, it):
		self._n = n
		self._it = it
		self._fill_cache()
	
	def __iter__(self):
		return self

	def _fill_cache(self):
		"""fills cache with up to `self._n` items"""
		self._cache = []
		self._cache_head = 0

		assert self._n > 0
		n = 0
		try:
			while n < self._n:
				self._cache.append(next(self._it))
				n += 1
		except StopIteration: pass

	def _rotate(self, item):
		"""add the given item to the end of the cache, pushing off (and returning) the current head"""
		try:
			old_item = self._cache[self._cache_head]
		except IndexError as e:
			print("Key Error %s" % (e,))
			print("cache = %r" % (self._cache))
			print("cache head = %r" % (self._cache_head))
			raise
		self._cache[self._cache_head] = item
		self._cache_head = (self._cache_head + 1) % self._n
		return old_item
	@property
	def end(self):
		exhaust(self)
		return self._cache[self._cache_head:] + self._cache[:self._cache_head]

	def __next__(self):
		return self._rotate(next(self._it))
	next = __next__ # for python2


def exhaust(iter):
	try:
		while True:
			next(iter)
	except StopIteration:
		pass

def iter_length(iter):
	"""
	>>> iter_length(iter([1,2,3]))
	3
	>>> iter_length(iter([]))
	0
	"""
	n = 0
	try:
		while True:
			next(iter)
			n+=1
	except StopIteration:
		return n

def drop(n, it, stop_short=False):
	"""
	>>> list(drop(3, iter([1,2,3,4,5])))
	[4, 5]
	>>> list(drop(10, iter([1,2,3,4,5]), True))
	[]
	"""
	try:
		while n>0:
			next(it)
			n-=1
		return it
	except StopIteration:
		if stop_short:
			return iter([])
		raise

def take(n, it, stop_short=False):
	"""
	>>> list(take(3, iter([1,2,3,4,5])))
	[1, 2, 3]
	>>> list(take(0, iter([1,2,3,4,5])))
	[]
	>>> list(take(3, iter([1]), True))
	[1]
	"""
	results = []
	try:
		while n>0:
			results.append(next(it))
			n-=1
		return results
	except StopIteration:
		if stop_short:
			return results
		raise



from __future__ import print_function
import subprocess
from piep.error import Exit
from piep.line import Line

active_commands = []
class Command(object):
	def __init__(self, cmd, check=None, **kw):
		self.cmd = cmd

		self.raise_on_error = check
		defaults = dict(stdout = subprocess.PIPE)
		defaults.update(kw)
		self.kwargs = defaults

		active_commands.append(self)
		self.checked = False
		self.proc = None
		self.stdout = None
		self.stderr = None
	
	def _spawn(self):
		if self.proc is None:
			try:
				self.proc = subprocess.Popen(self.cmd, **self.kwargs)
			except OSError as e:
				raise Exit("error executing %r: %s" % (list(self.cmd), e))
	
	def wait(self, raise_on_error=True):
		self._spawn()
		self.checked = True
		(self.stdout, self.stderr) = self.proc.communicate()
		self.ended = True
		self.status = self.proc.returncode
		self.succeeded = self.status == 0
		explicitly_suppressed = self.raise_on_error is False
		if self.raise_on_error or (raise_on_error and not explicitly_suppressed):
			if not self.succeeded:
				raise subprocess.CalledProcessError(self.status, ' '.join(self.cmd))
	
	def __bool__(self):
		self.wait(raise_on_error = False)
		return self.succeeded
	__nonzero__ = __bool__ # py2 compat

	def __str__(self):
		if not self.checked:
			self.wait()
		return (self.stdout.decode('utf-8') or '').rstrip('\n\r')
	str = property(__str__)
	def __repr__(self): return repr(str(self))
	def __add__(self, other): return str(self).__add__(other)
	def __radd__(self, other): return str(self).__radd__(other)
	def __contains__(self, other): return str(self).__contains__(other)
	def __eq__(self, other): return str(self).__eq__(other)
	def __hash__(self): return str(self).__hash__()
	def __ne__(self, other): return str(self).__ne__(other)
	def __ge__(self, other): return str(self).__ge__(other)
	def __gt__(self, other): return str(self).__gt__(other)
	def __lt__(self, other): return str(self).__lt__(other)
	def __le__(self, other): return str(self).__le__(other)
	def __mod__(self, other): return str(self).__mod__(other)
	def __rmod__(self, other): return str(self).__rmod__(other)
	def __mul__(self, other): return str(self).__mul__(other)
	def __rmul__(self, other): return str(self).__rmul__(other)
	def __getitem__(self, other): return str(self).__getitem__(other)

	def __getattr__(self, attr):
		return getattr(Line(self), attr)
	
def check_for_failed_commands():
	global active_commands
	try:
		for cmd in active_commands:
			if not cmd.checked:
				cmd.wait()
	finally:
		active_commands = []




import sys
is_python_2 = sys.version_info[0] == 2
def execfn(a,b):
	exec(a,b)

if is_python_2:
	import itertools
	map = itertools.imap
	filter = itertools.ifilter
	zip = itertools.izip
	zip_longest = itertools.izip_longest
	import __builtin__ as python_builtin
else:
	basestring = str
	import builtins as python_builtin



#!/usr/bin/env python3

from __future__ import print_function
import os, sys
import ast
from optparse import OptionParser
import itertools

from piep.sequence import Stream, BaseList, List
from piep.line import Line
from piep.builtins import builtins
from piep.error import Exit

from .pycompat import *

DEBUG = False
def debug(*a):
	if not DEBUG: return
	print(*a)

def main(argv=None):
	if argv is None:
		argv = sys.argv[1:]
	try:
		opts, args = parse_args(argv)
		lines = run(opts, args)
		print_results(lines, opts)
		return 0
	except Exit as e:
		if e.message:
			print(e.message, file=sys.stderr)
		sys.exit(1)
	except KeyboardInterrupt:
		print("interrupted", file=sys.stderr)
		sys.exit(1)
	except Exception as e:
		if DEBUG:
			import traceback
			traceback.print_exc(file=sys.stderr)
		if isinstance(e, (AttributeError, KeyError, AssertionError, IndexError)):
			# these ones don't have very good error messages:
			e = "%s: %s" % (type(e).__name__, e)
		print(str(e), file=sys.stderr)
		sys.exit(1)

def parse_args(argv=None):
	global DEBUG
	if argv is None: argv = sys.argv[1:]
	p = OptionParser('usage: piep [OPTIONS] PIPELINE')
	p.add_option('--debug', action='store_true')
	p.add_option('-j', '--join', default=' ')
	p.add_option('-e', '--eval', action='append', dest='evals', default=[], metavar='EVAL', help='evaluate arbitrary code before running the script (in global scope, may be given multiple times)')
	p.add_option('-m', '--import', action='append', dest='imports', default=[], metavar='MODULE', help='add a module to global scope (may be given multiple times)')
	p.add_option('-f', '--file', action='append', dest='files', default=[], metavar='FILE', help='add another input stream (available as files[n])')
	p.add_option('-i', '--input', dest='input', help='use a named file (instead of stdin)')
	p.add_option('-p', '--path', action='append', dest='import_paths', default=[], help='add a location to the import path (the same as $PYTHONPATH / sys.path)')
	p.add_option('-0', '--read0', action='store_true', dest='input_nullsep', help='read input as null-separated fields')
	p.add_option('-n', '--no-input', action='store_true', help='don\'t read stdin - self-constructing pipeline')
	p.add_option('--print0', action='store_true', dest='output_nullsep', help='print output as null-separated fields')
	opts, args = p.parse_args(argv)
	assert len(args) > 0, "Not enough arguments\n" + p.format_help()
	assert len(args) == 1, "Too many arguments\n" + p.format_help()
	DEBUG = opts.debug
	return (opts, args)

def print_results(lines, opts):
	'''print lines separated by either \n or \0'''
	lines = iter(lines)
	print_line = print
	if opts.output_nullsep:
		try:
			line = next(lines)
		except StopIteration:
			return
		print(line, end='')
		def print_line(line):
			print('\0', line, end='', sep='')
	for line in lines:
		print_line(line)

def run(opts, args):
	cmd = args[0]
	input_file = open(opts.input) if opts.input else sys.stdin

	opts.join = opts.join.encode('utf-8').decode('unicode_escape')

	bindings = init_globals(opts, input_file)

	debug("Pipeline string: %s" % (cmd,))
	exprs = split_on_pipes(cmd)
	debug("Split expressions:\n  - " + "\n  - ".join(exprs))
	assert len(exprs) > 0, "You must provide at least one expression"
	if opts.no_input:
		first_expr = parse_expr(exprs[0])
		if isinstance(first_expr, ast.Assign):
			assigned_names = [name.id for name in first_expr.targets]
			assert 'pp' in assigned_names, "The first expression must assign to `pp` when --no-input is specified"
		else:
			exprs[0] = 'pp=' + exprs[0]
	output = eval_pipes(exprs, bindings)

	# strings are iterable, but we don't want to do that!
	if isinstance(output, basestring):
		output = [output]
	try:
		output = iter(output)
	except TypeError as err:
		debug(err)
		yield output
	else:
		for line in output:
			if line is not None:
				if isinstance(line, (tuple, list)):
					line = opts.join.join(map(str, line))
				yield line

def init_globals(opts, input_file):
	def make_stream(f):
		return Stream(map(lambda x: Line(x.rstrip('\n\r')), iter(f)))

	pp = make_stream(input_file)
	globs = builtins.copy()
	globs['pp'] = pp
	for path in opts.import_paths:
		path = os.path.abspath(path)
		if path not in sys.path:
			sys.path.insert(0, path)
	for import_mod in opts.imports:
		import_node = ast.Import(names=[ast.alias(name=import_mod, asname=None)])
		code = compile(ast.fix_missing_locations(ast.Module(body=[import_node], type_ignores=[])), 'import %s' % (import_mod,), 'exec')
		eval(code, globs)
	for eval_str in opts.evals:
		try:
			execfn(eval_str, globs)
		except SyntaxError as e:
			raise Exit("got error: %s\nwhile evaluating: %s" % (e,eval_str))
	files = [make_stream(open(f)) for f in opts.files]
	globs['files'] = files
	if len(files) > 0: # convenience for single file operation
		globs['ff'] = files[0]
	return globs


def split_on_pipes(cmds):
	r'''
	splits total command array based on pipes taking into account quotes,
	parentheses and escapes. returns array of commands that will be processed procedurally.

	>>> split_on_pipes("a | b | c")
	['a', 'b', 'c']

	>>> split_on_pipes(r"a | (b|'\\') | d")
	['a', "(b|'\\\\')", 'd']

	>>> split_on_pipes("""a | b'"' | d'"'""")
	['a', 'b\'"\'', 'd\'"\'']

	>>> split_on_pipes('a | "(b|c" | d')
	['a', '"(b|c"', 'd']

	>>> split_on_pipes('a[1|2] | b')
	['a[1|2]', 'b']

	>>> split_on_pipes('{"x": 1|2} | b')
	['{"x": 1|2}', 'b']

	>>> split_on_pipes('{"x": (1)|2} | b')
	['{"x": (1)|2}', 'b']

	>>> split_on_pipes(r'a.replace("/", "\\") | b')
	['a.replace("/", "\\\\")', 'b']
	'''
	
	OPENERS = frozenset(('{', '(', '['))
	CLOSERS = {
		'}':'{',
		')':'(',
		']':'[',
	}
	QUOTES = frozenset(('"',"'"))

	cmd_array = []
	letters = list(cmds)

	cmd = ''
	escape = False
	context = []
	while letters:
		letter = letters.pop(0)
		open_ctx = context[-1] if context else None
		debug("got letter %r, open_ctx = %r, esc=%s" % (letter, open_ctx, escape))

		if not escape:
			# behave differently if inside quotes
			if open_ctx in QUOTES:
				if open_ctx == letter:
					context.pop()
			else:
				# quotes and brackets can nest in anything but quotes
				if letter in QUOTES or letter in OPENERS:
					context.append(letter)
				elif letter in CLOSERS:
					opener = CLOSERS[letter]
					if open_ctx == opener:
						context.pop()

			if letter == '|' and open_ctx is None:
				cmd_array.append(cmd.strip())
				cmd = ''
				continue

		cmd += letter

		# if backslash, set `escape` for the next letter we encounter,
		# note that two backslashes in a row reverts to unescaped
		escape = letter == '\\' and not escape

	cmd_array.append(cmd.strip())
	return cmd_array

class Mode(object):
	__slots__ = ['desc', 'vars']
	def __init__(self, desc, vars):
		self.desc = desc
		self.vars = vars
	def __str__(self): return self.desc
	def __repr__(self): return "<# mode: %s>" % (self,)

class Modes(object):
	__slots__ = ['GLOBAL', 'LINE']
	def __init__(self):
		self.GLOBAL = Mode('global', set(['pp','files','ff']))
		self.LINE = Mode('line', set(['p', 'i']))
MODE = Modes()
RESERVED_VARS = MODE.GLOBAL.vars.union(MODE.LINE.vars)
NON_ASSIGNABLE_VARS = RESERVED_VARS.difference(['pp', 'p'])

def detect_mode(expr, source_text):
	names = set()
	class NameFinder(ast.NodeVisitor):
		def visit_Assign(self, node):
			assigned_names = set(name.id for name in node.targets if isinstance(name, ast.Name))
			reserved_names = list(assigned_names.intersection(NON_ASSIGNABLE_VARS))
			if reserved_names:
				raise AssertionError("can't assign to `%s` (expression: %s)" % (sorted(reserved_names)[0], source_text))
			names.update(assigned_names)
			self.generic_visit(node)

		def visit_Name(self, node):
			if isinstance(node.ctx, ast.Load):
				names.add(node.id)
			self.generic_visit(node)
	NameFinder().visit(expr)
	debug('expr references: %r' % (names,))

	mode = None
	if len(names.intersection(MODE.GLOBAL.vars)) > 0:
		mode = MODE.GLOBAL

	if len(names.intersection(MODE.LINE.vars)) > 0:
		assert mode is None, "Can't use file-level and line-level expressions in the same pipe expression: " + source_text
		mode = MODE.LINE
	return (mode or MODE.LINE), names

def parse_expr(expr):
	try:
		return ast.parse(expr + '\n', mode='eval').body
	except SyntaxError as e:
		try:
			stmt = ast.parse(expr + '\n', mode='exec').body
			if len(stmt) == 1 and isinstance(stmt[0], ast.Assign):
				# single assignments are OK
				return stmt[0]
			else:
				raise e
		except SyntaxError as e:
			raise Exit("got error: %s\nwhile evaluating: %s" % (e,expr))

def compile_pipe_exprs(exprs):
	body = []

	# this code doesn't need to be parameterised, so we'll just parse a string
	post_pipe_check = ast.parse(
			"_check_for_failed_commands()\n"
		).body
	post_expr_check = post_pipe_check + ast.parse(
			"if callable(_p): _p = _p(p)\n"
			"p = p if _p is True else (None if _p is False else _p)\n"
			"if p is None: return None\n"
		).body

	def annotate(expr):
		'''get the mode & referenced variables for a given ast node'''
		ast_node = parse_expr(expr)
		mode, vars = detect_mode(ast_node, expr)
		return ast_node, mode, vars

	def name(id, ctx=None):
		if ctx is None:
			ctx = ast.Load()
		return ast.Name(id=id, ctx=ctx)

	if is_python_2:
		arg = name
	else:
		def arg(id):
			return ast.arg(id, None)

	def assign(var, expr):
		return ast.Assign(targets=[name(var, ctx=ast.Store())], value=expr)

	def combine_pipe_transforms(body):
		'''
		Takes a list of expressions and creates a list of statements
		to process the entirety of ``pp`` with the given sub-pipeline.
		Importantly, the expressions exist in the same lexical scope, so
		assignments are available for use in subsequent expressions.
		'''
		transform_def = ast.FunctionDef(
			name='_transformer',
			args=ast.arguments(
				posonlyargs=[],
				args=[arg('p'), arg('i')],
				vararg=None,
				kwarg=None,
				kwonlyargs=[],
				kw_defaults=[],
				defaults=[]
			),
			body=body + [ast.Return(name('p'))],
			decorator_list=[])

		map_fn = ast.Attribute(value=name('pp'), attr='map_index', ctx=ast.Load())
		assign_pp = assign('pp', ast.Call(
				func=map_fn,
				args=[name('_transformer')],
				keywords=[],
				starargs=None,
				kwargs=None
				))
		return [
			transform_def,
			assign_pp
			]

	annotated_exprs = map(annotate, exprs)
	is_linewise = lambda x: x[1] is MODE.LINE

	def ensure_stream():
		call = ast.Call(
			func=name('_ensure_piep_sequence'),
			args=[name('pp')],
			keywords=[],
			starargs=None,
			kwargs=None
		)
		body.append(assign('pp', call))

	for linewise, group in itertools.groupby(annotated_exprs, is_linewise):
		if linewise:
			group_body = []
			for item in group:
				expr, mode, vars = item
				if not isinstance(expr, ast.Assign):
					group_body.append(assign('_p', expr))
					group_body.extend(post_expr_check)
				else:
					group_body.append(expr)
			ensure_stream()
			body.extend(combine_pipe_transforms(group_body))
		else:
			for item in group:
				expr = item[0]
				ensure_stream()
				if not isinstance(expr, ast.Assign):
					expr = assign('pp', expr)
				body.append(expr)
				body.extend(post_pipe_check)
	
	mod = ast.Module(body=body, type_ignores=[])
	ast.fix_missing_locations(mod)

	#raise RuntimeError(ast.dump(mod))
	return compile(mod, '(input)', 'exec')

def eval_pipes(exprs, bindings):
	mod = compile_pipe_exprs(exprs)
	execfn(mod, bindings)
	return bindings['pp']

if __name__ == '__main__':
	main()



'''
Standard modules:

.. data:: re
.. data:: os
.. data:: sys

Aliases:

.. data:: path

	Alias for the ``os.path`` module

.. data:: Line

	Alias for the ``piep.Line`` class (a subclass of ``str`` containing all the same methods as ``p`` does)

.. data:: List

	Alias for the ``piep.List`` class (a subclass of ``list`` containing all the same methods as ``pp`` does from :class:`piep.BaseList`)

.. data:: devnull

	A readable and writable file pointing to the null device (/dev/null)

Globally-accessible functions:

'''
from __future__ import print_function
import subprocess
import re, os, sys
from .pycompat import *

from piep.sequence import iter_length, BaseList, List, Stream
from piep import line
from piep.shell import Command, check_for_failed_commands
builtins = {}


def add_builtin(fn, name=None):
	name = name or fn.__name__
	if name in builtins:
		print('WARN: overriding piep_builtin %r' % (name,))
	builtins[name] = fn
	return fn

add_builtin(re, 're')
add_builtin(os, 'os')
add_builtin(os.path, 'path')
add_builtin(sys, 'sys')
add_builtin(line.Line, 'Line')
add_builtin(List, 'List')

@add_builtin
def len(obj):
	'''like the builtin ``len``, but works (destructively) on iterators.'''
	try:
		return python_builtin.len(obj)
	except TypeError as e:
		try:
			it = iter(obj)
		except (TypeError, AttributeError) as e2:
			raise e
		return iter_length(it)

@add_builtin
def str(obj):
	'''like the builtin ``str``, but returns an instance of the ``piep.Line`` subclass.'''
	return line.Line(obj)

@add_builtin
def sh(*args, **kwargs):
	'''
	Invoke a shell program and return its output. ``*args`` will be the program name + arguments, and
	``**kwargs`` will be passed through to ``subprocess.Popen``.

	One additional keyword argument is supported - the ``check`` keyword argument (used to suppress an exception when the command fails).

	For more info, see :ref:`running shell commands`
	'''
	return Command(args, **kwargs)

@add_builtin
def spawn(*a, **k):
	'''acts exactly like ``sh``, except that it just returns ``True``.

	For more info, see :ref:`running shell commands`
	'''

	sh(*a, **k)
	return True

devnull = open(os.devnull, 'r+')
add_builtin(devnull, 'devnull')

@add_builtin
@line.wrap_multi
def shellsplit(s):
	'''Split a string much like a shell would.
	(alias for ``shlex.split``)
	'''
	import shlex
	return shlex.split(s)

@add_builtin
def pretty(obj):
	'''
	return a console-cloured pretty printed version of ``obj``
	(like ``repr(obj)``, but coloured)
	'''
	import pygments
	import pygments.lexers
	import pygments.formatters
	return pygments.highlight(repr(obj), lexer=pygments.lexers.get_lexer_by_name('py'), formatter=pygments.formatters.get_formatter_by_name('console')).rstrip('\n\r')

@add_builtin
def _ensure_piep_sequence(pp):
	if not isinstance(pp, BaseList):
		if isinstance(pp, basestring):
			pp = pp.splitlines()
		# if the last expr turned pp into a normal list or some other iterable, fix that...
		cls = Stream
		if isinstance(pp, list):
			cls = List
		try:
			pp = cls(iter(pp))
		except TypeError: # pp was presumably not iterable
			pass
	return pp


@add_builtin
def ignore(*a):
	'''Ignore all arguments and returns ``True``.
	Occasionally useful to evaluate an expression for its
	side-effects without making use of its value.'''
	return True

add_builtin(check_for_failed_commands, "_check_for_failed_commands")



from __future__ import print_function
import os
import re

def _apply_doc(ret, fn, doc):
	ret.__name__ = fn.__name__
	if doc is True:
		ret.__doc__ = fn.__doc__
	elif doc is not False:
		ret.__doc__ = doc

def wrap(fn, doc=False):
	ret = lambda *a, **k: Line(fn(*a, **k))
	_apply_doc(ret, fn, doc)
	return ret

def wrap_multi(fn, doc=False):
	def ret(*a, **k):
		from piep.sequence import List
		return List(Line(l) for l in fn(*a, **k))
	_apply_doc(ret, fn, doc)
	return ret

class Line(str):
	def __new__(cls, s, *a, **k):
		if s is None:
			return s
		return super(Line, cls).__new__(cls, s, *a, **k)

	@classmethod
	def add_method(cls, fn, name = None):
		setattr(cls, name or fn.__name__, fn)

	@wrap
	def ext(s):
		'''Return the filename extension (including ".")'''
		return os.path.splitext(s)[1]

	@wrap
	def reversed(s):
		'''Return a reversed version of this string'''
		return s[::-1]

	@wrap
	def extonly(s):
		'''
		Return the filename extension (without the ".")
		If there is no file extension, returns ``None``
		'''
		ext = os.path.splitext(s)[1]
		return ext[1:] if ext else None

	splitext = wrap_multi(os.path.splitext, doc='alias for os.path.splitext(self)')
	dirname = wrap(os.path.dirname, doc='alias for os.path.dirname(self)')
	basename = wrap(os.path.basename, doc='alias for os.path.basename(self)')
	filename = basename
	stripext = wrap(lambda s: os.path.splitext(s)[0], doc='remove filetype extension if present, including "."')
	splitline = wrap_multi(str.splitlines, doc='alias for str.splitlines()')
	splittab = wrap_multi(lambda s: s.split('\t'), doc='split on "\\t"')
	splitcomma = wrap_multi(lambda s: s.split(','), doc='split on ","')
	splitcolon = wrap_multi(lambda s: s.split(':'), doc='split on ":"')
	splitslash = wrap_multi(lambda s: s.split('/'), doc='split on "/"')
	splitpath = wrap_multi(lambda s: s.split(os.path.sep), doc='split on path separator ("/" for unix, "\\" for windows)')

	@wrap_multi
	def shellsplit(self):
		'''shlex.split(self)'''
		import shlex
		return shlex.split(self)

	@wrap_multi
	def splitre(self, regex, *a, **kw):
		'''re.split(regex, self, *a, **kw)'''
		return re.split(regex, self, *a, **kw)

	@wrap
	def match(self, pattern, group=0, flags=0):
		'''extract matched part of the string (or a captured group, if ``group`` is given)'''
		res = re.search(pattern, self, flags=flags)
		return res and res.group(group or 0)

	def matches(self, pattern, group=0, flags=0):
		'''return True or False depending on if the given regex can be found anywhere in the line'''
		return bool(re.search(pattern, self, flags=flags))

	# copy all `str` builtins
	capitalize = wrap(str.capitalize)
	center     = wrap(str.center)
	expandtabs = wrap(str.expandtabs)
	format     = wrap(str.format)
	join       = wrap(str.join)
	ljust      = wrap(str.ljust)
	lower      = wrap(str.lower)
	lstrip     = wrap(str.lstrip)
	partition  = wrap_multi(str.partition)
	replace    = wrap(str.replace)
	rjust      = wrap(str.rjust)
	rpartition = wrap_multi(str.rpartition)
	rsplit     = wrap_multi(str.rsplit)
	rstrip     = wrap(str.rstrip)
	split      = wrap_multi(str.split)
	splitlines = wrap_multi(str.splitlines)
	strip      = wrap(str.strip)
	swapcase   = wrap(str.swapcase)
	title      = wrap(str.title)
	translate  = wrap(str.translate)
	upper      = wrap(str.upper)
	zfill      = wrap(str.zfill)

	# and specials:
	__add__      = wrap(str.__add__)
	__format__   = wrap(str.__format__)
	__getitem__  = wrap(str.__getitem__)
	__mod__      = wrap(str.__mod__)
	__mul__      = wrap(str.__mul__)
	__rmod__     = wrap(str.__rmod__)
	__rmul__     = wrap(str.__rmul__)

